import logging
import os
from typing import Dict

import numpy as np

from .device_util import GPU_DEVICE, DEFAULT_FP16_VALUE, DEFAULT_WORKSPACE_VALUE, DEFAULT_BATCH_SIZE


class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        """Within this context, host_mom means the cpu memory and device means the GPU memory
        """
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


class TensorRTServable:
    def __init__(self, local_path: str):
        self.local_path = local_path
        self.model_ins = None
        self.model_init_stat = False
        self.device = GPU_DEVICE

        self.fp16 = DEFAULT_FP16_VALUE
        self.workspace = DEFAULT_WORKSPACE_VALUE
        self.max_batch_size = DEFAULT_BATCH_SIZE

        self.optim_batch_size = self.max_batch_size

        self.model_input_list = None
        self.model_output_list = None

        self.engine_file = self.local_path + ".engine"
        self.trt_logger = None
        self.engine = None
        self.context = None
        self.bindings = None
        self.input_bindings = None
        self.output_bindings = None
        self.binding_idx = None
        self.stream = None

    def load(self):
        if self.model_init_stat:
            return

        import pycuda.driver as cuda
        import pycuda.autoinit  # pylint: disable=unused-import
        import tensorrt as trt

        self.trt_logger = trt.Logger(trt.Logger.Severity.INFO)

        if os.path.exists(self.engine_file):
            with open(self.engine_file, "rb") as f_engine, \
                    trt.Runtime(self.trt_logger) as runtime:

                self.engine = runtime.deserialize_cuda_engine(f_engine.read())
        else:
            network_creation_flag = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)

            with trt.Builder(self.trt_logger) as builder, \
                    builder.create_network(network_creation_flag) as network, \
                    trt.OnnxParser(network, self.trt_logger) as parser:

                builder.max_workspace_size = self.workspace
                builder.max_batch_size = 1
                builder.fp16_mode = self.fp16

                # parse onnx file
                with open(self.local_path, "rb") as f_onnx:
                    if not parser.parse(f_onnx.read()):
                        logging.error("onnx file parse error")
                        for i in range(parser.num_errors):
                            logging.error("error %d: %s", i, parser.get_error(i))

                        raise Exception("onnx file parse error")

                # set dynamic batch size
                builder_config = builder.create_builder_config()
                builder_config.max_workspace_size = self.workspace

                if self.fp16:
                    builder_config.flags = 1 << trt.BuilderFlag.FP16

                profile = builder.create_optimization_profile()
                for i in range(network.num_inputs):
                    input_name = network.get_input(i).name
                    input_shape = list(network.get_input(i).shape)
                    min_shape = input_shape.copy()
                    min_shape[0] = 1
                    optim_shape = input_shape.copy()
                    optim_shape[0] = self.optim_batch_size
                    max_shape = input_shape.copy()
                    max_shape[0] = self.max_batch_size
                    profile.set_shape(input_name, min_shape, optim_shape, max_shape)

                builder_config.add_optimization_profile(profile)
                self.engine = builder.build_engine(network, config=builder_config)

                with open(self.engine_file, "wb") as f_engine:
                    f_engine.write(self.engine.serialize())

        self.context = self.engine.create_execution_context()

        # allocate buffers
        self.model_input_list = []
        self.model_output_list = []
        self.input_bindings = {}
        self.output_bindings = {}
        self.bindings = []
        self.binding_idx = {}

        for i, binding in enumerate(self.engine):
            max_shape = list(self.engine.get_binding_shape(binding))
            max_shape[0] = self.max_batch_size
            size = trt.volume(max_shape)
            dtype = trt.nptype(self.engine.get_binding_dtype(binding))
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            # Append the device buffer to device bindings.
            self.bindings.append(int(device_mem))
            # Append to the appropriate list.
            if self.engine.binding_is_input(binding):
                self.input_bindings[binding] = HostDeviceMem(host_mem, device_mem)
                self.model_input_list.append(binding)
            else:
                self.output_bindings[binding] = HostDeviceMem(host_mem, device_mem)
                self.model_output_list.append(binding)

            self.binding_idx[binding] = i

        self.stream = cuda.Stream()

    def forward(self, inputs: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        import pycuda.driver as cuda
        import tensorrt as trt

        input_dict = {}
        for key in self.model_input_list:
            if key not in inputs:
                raise Exception("necessary input (%s) missing and model needs %s" % (key, self.model_input_list))
            input_dict[key] = inputs[key]

        batch_size = next(iter(inputs.values())).shape[0]

        # set input binding shape
        for input_name, arr in inputs.items():
            self.context.set_binding_shape(self.binding_idx[input_name], arr.shape)

        # transfer data to GPU
        for input_name, arr in inputs.items():
            np.copyto(self.input_bindings[input_name].host[:arr.size], np.ravel(arr))
            cuda.memcpy_htod_async(self.input_bindings[input_name].device,
                                   self.input_bindings[input_name].host[:arr.size],
                                   stream=self.stream)

        # inference
        self.context.execute_async_v2(self.bindings, self.stream.handle)

        # transfer data back to CPU
        for output_name, binding in self.output_bindings.items():
            shape = list(self.engine.get_binding_shape(output_name))
            shape[0] = batch_size

            size = trt.volume(shape)
            cuda.memcpy_dtoh_async(binding.host[:size], binding.device, stream=self.stream)

        self.stream.synchronize()

        output_dict = {}
        for output_name in self.model_output_list:
            shape = list(self.engine.get_binding_shape(output_name))
            shape[0] = batch_size

            size = trt.volume(shape)
            arr = self.output_bindings[output_name].host[:size].copy()
            arr = np.reshape(arr, shape)
            output_dict[output_name] = arr

        return output_dict


import logging
from typing import Dict

import numpy as np

from .device_util import CPU_DEVICE, get_cpu_procs, get_gpu_available

PROVIDER_LIST = ["TensorrtExecutionProvider", "CUDAExecutionProvider",
                 "OpenVINOExecutionProvider", "CPUExecutionProvider"]


class ONNXServable:
    """
    onnx servable 输入dict[str, ndarray]，输出dict[str, ndarray]
    """

    def __init__(self, local_path: str):
        self.local_path = local_path
        self.model_ins = None
        self.model_init_stat = False

        self.device = CPU_DEVICE
        self.provider_key = None

        self.model_input_list = None
        self.model_output_list = None

    def load(self):
        if not self.model_init_stat:
            import onnxruntime
            logging.info("onnxruntime current device: %s, available execution providers: %s",
                         onnxruntime.get_device(), str(onnxruntime.get_available_providers()))

            options = onnxruntime.SessionOptions()
            options.execution_mode = onnxruntime.ExecutionMode.ORT_SEQUENTIAL
            options.inter_op_num_threads = 1
            options.intra_op_num_threads = get_cpu_procs()
            options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL

            if self.provider_key is None:
                providers = onnxruntime.get_available_providers()
            else:
                providers = self.provider_key.copy()

            if not get_gpu_available() or self.device == CPU_DEVICE:
                if "CUDAExecutionProvider" in providers:
                    providers.remove("CUDAExecutionProvider")
                if "TensorrtExecutionProvider" in providers:
                    providers.remove("TensorrtExecutionProvider")

            ordered_providers = []
            for provider in PROVIDER_LIST:
                if provider in providers:
                    ordered_providers.append(provider)
            logging.info("use execution providers: %s", str(ordered_providers))

            self.model_ins = onnxruntime.InferenceSession(self.local_path, options, providers=ordered_providers)
            self.model_init_stat = True

            self.model_input_list = []
            for input_obj in self.model_ins.get_inputs():
                self.model_input_list.append(input_obj.name)
                logging.info("model input name: %s shape: %s", input_obj.name, input_obj.shape)

            self.model_output_list = []
            for output_obj in self.model_ins.get_outputs():
                self.model_output_list.append(output_obj.name)
                logging.info("model output name: %s shape: %s", output_obj.name, output_obj.shape)

    def forward(self, inputs: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        if not isinstance(inputs, dict):
            raise Exception("onnx servable can only support dict input, got (%s) type" % type(inputs))

        input_dict = {}
        for key in self.model_input_list:
            if key not in inputs:
                raise Exception("necessary input (%s) missing and model needs %s" % (key, self.model_input_list))
            input_dict[key] = inputs[key]

        output_list = self.model_ins.run(None, input_dict)
        output_dict = {}
        for i, key in enumerate(self.model_output_list):
            output_dict[key] = output_list[i]

        return output_dict



import logging
import os
from typing import Dict

import numpy as np

from .device_util import CPU_DEVICE, get_cpu_procs


class OpenvinoServable:
    """
    openvino servable 输入dict[str, ndarray]，输出dict[str, ndarray]
    """

    def __init__(self, local_path: str):
        self.local_path = local_path
        self.model_ins = None
        self.model_init_stat = False

        self.device = CPU_DEVICE

        self.model_input_list = None

    def load(self):
        if not self.model_init_stat:
            from openvino.inference_engine import IECore

            # --------------------------- 1. Read IR Generated by ModelOptimizer (.xml and .bin files) ------------
            ie_core = IECore()
            model_xml = self.local_path
            model_bin = os.path.splitext(model_xml)[0] + ".bin"
            net = ie_core.read_network(model=model_xml, weights=model_bin)
            # -----------------------------------------------------------------------------------------------------

            # ------------- 2. Load Plugin for inference engine and extensions library if specified --------------
            supported_layers = ie_core.query_network(net, self.device)
            not_supported_layers = [l for l in net.layers.keys() if l not in supported_layers]
            if len(not_supported_layers) != 0:
                logging.error("Following layers are not supported by the plugin for specified device %s: %s",
                              self.device, ', '.join(not_supported_layers))
                exit()
            # -----------------------------------------------------------------------------------------------------

            # ---------------------- 3. instantiate an executable version of the model ----------------------------
            available_procs = str(get_cpu_procs())
            self.model_ins = ie_core.load_network(network=net,
                                                  device_name=self.device,
                                                  config={"CPU_THREADS_NUM": available_procs})
            self.model_init_stat = True
            # -----------------------------------------------------------------------------------------------------

            # --------------------------- 4. Configure input & output ---------------------------------------------
            self.model_input_list = []
            for input_obj_name in self.model_ins.inputs:
                self.model_input_list.append(input_obj_name)
                logging.info("model input name: %s shape: %s",
                             input_obj_name, self.model_ins.inputs[input_obj_name].shape)
            # -----------------------------------------------------------------------------------------------------

    def forward(self, inputs: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        if not isinstance(inputs, dict):
            raise Exception("openvino servable can only support dict input, got (%s) type" % str(type(inputs)))

        input_dict = {}
        for key in self.model_input_list:
            if key not in inputs:
                raise Exception("necessary input (%s) missing and model needs %s" % (key, self.model_input_list))
            input_dict[key] = inputs[key]

        output_dict = self.model_ins.infer(input_dict)

        return output_dict



from .device_util import CPU_DEVICE, get_cpu_procs, get_gpu_available
from tml_serving.utils.torch_util import place_on_device, set_torch_num_threads

class LibPytorchServable:
    """
    libpytorch servable 输入dict[str, ndarray]，输出dict[str, ndarray]
    """

    def __init__(self, local_path: str):
        self.local_path = local_path
        self.model_ins = None
        self.model_init_stat = False

        self.device = CPU_DEVICE
        self.provider_key = None

        self.model_input_list = None
        self.model_output_list = None

    def load(self):
        if not self.model_init_stat:
            import torch
            set_torch_num_threads()

            if get_gpu_available() and self.device != CPU_DEVICE:
                self.torch_device = torch.device(GPU_DEVICE)
            else:
                self.torch_device = torch.device(CPU_DEVICE)

            self.model_ins = torch.jit.load(self.local_path, map_location=self.torch_device)
            self.model_ins.eval()
            self.model_init_stat = True

    def forward(self, *args, **kwargs):
        import torch
        # 默认实现，torch模型返回值是tensor
        args = place_on_device(args, self.torch_device)
        kwargs = place_on_device(kwargs, self.torch_device)

        with torch.no_grad():
            if len(args) == 1 and isinstance(args[0], tuple):
                output = self.model_ins(*args[0], **kwargs)
            else:
                output = self.model_ins(*args, **kwargs)

            output = place_on_device(output, torch.device(CPU_DEVICE))
            return output



GPU_DEVICE = "cuda"
CPU_DEVICE = "cpu"

DEFAULT_FP16_VALUE = False
DEFAULT_WORKSPACE_VALUE = 12 * 1 << 30  # 12GB
DEFAULT_BATCH_SIZE = 4


def get_cpu_procs():
    import automaxprocs
    available_procs = automaxprocs.cpu_count()
    max_procs = automaxprocs.get_max_procs()
    if max_procs != -1:
        available_procs = max_procs
    return available_procs


def get_gpu_available():
    import pynvml
    try:
        pynvml.nvmlInit()
        device_count = pynvml.nvmlDeviceGetCount()
        return device_count > 0
    except:  # pylint: disable=bare-except
        return False



import os
import re

from setuptools import find_packages
from setuptools import setup


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__), "ronda_servable", "__init__.py")
    with open(init_py) as f_in:
        for line in f_in:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        raise RuntimeError("Cannot find version in {}".format(init_py))


setup(
    name="ronda_servable",
    version=read_version(),
    url="https://git.woa.com/RondaServing/RondaServable/RondaPythonServable",
    project_urls={
        "Documentation": "https://git.woa.com/RondaServing/RondaServable/RondaPythonServable/wikis/home",
        "Code": "https://git.woa.com/RondaServing/RondaServable/RondaPythonServable",
        "Issue tracker": "https://git.woa.com/RondaServing/RondaServable/RondaPythonServable/issues",
    },
    maintainer="Ronda Serving Group",
    description="Ronda Python Servable",
    python_requires=">=3.6",
    install_requires=[
        "numpy",
        "automaxprocs",
        "pynvml"
    ],
    packages=find_packages(include=("ronda_servable", "ronda_servable.*")),
    package_data={'': ['*.so']},
)
